@@ -4,9 +4,29 @@
 /* eslint-env browser */
 "use strict";
 
-const { Component, createFactory } = require("devtools/client/shared/vendor/react");
+const {
+  Component,
+  createFactory,
+} = require("devtools/client/shared/vendor/react");
 const PropTypes = require("devtools/client/shared/vendor/react-prop-types");
 const dom = require("devtools/client/shared/vendor/react-dom-factories");
+const { scrollIntoView } = require("devtools/client/shared/scroll");
+const {
+  preventDefaultAndStopPropagation,
+} = require("devtools/client/shared/events");
+
+loader.lazyRequireGetter(
+  this,
+  "wrapMoveFocus",
+  "devtools/client/shared/focus",
+  true
+);
+loader.lazyRequireGetter(
+  this,
+  "getFocusableElements",
+  "devtools/client/shared/focus",
+  true
+);
 
 const AUTO_EXPAND_DEPTH = 0;
 const NUMBER_OF_OFFSCREEN_ITEMS = 1;
@@ -195,9 +215,19 @@ class Tree extends Component {
       // Handle when a new item is focused.
       onFocus: PropTypes.func,
 
+      // The currently active (keyboard) item, if any such item exists.
+      active: PropTypes.any,
+
       // Handle when item is activated with a keyboard (using Space or Enter)
       onActivate: PropTypes.func,
 
+      // The currently shown item, if any such item exists.
+      shown: PropTypes.any,
+
+      // Indicates if pressing ArrowRight key should only expand expandable node
+      // or if the selection should also move to the next node.
+      preventNavigationOnArrowRight: PropTypes.bool,
+
       // The depth to which we should automatically expand new items.
       autoExpandDepth: PropTypes.number,
 
@@ -228,6 +258,7 @@ class Tree extends Component {
   static get defaultProps() {
     return {
       autoExpandDepth: AUTO_EXPAND_DEPTH,
+      preventNavigationOnArrowRight: true,
     };
   }
 
@@ -238,7 +269,7 @@ class Tree extends Component {
       scroll: 0,
       height: window.innerHeight,
       seen: new Set(),
-      mouseDown: false
+      mouseDown: false,
     };
 
     this._onExpand = oncePerAnimationFrame(this._onExpand).bind(this);
@@ -246,25 +277,30 @@ class Tree extends Component {
     this._onScroll = oncePerAnimationFrame(this._onScroll).bind(this);
     this._focusPrevNode = oncePerAnimationFrame(this._focusPrevNode).bind(this);
     this._focusNextNode = oncePerAnimationFrame(this._focusNextNode).bind(this);
-    this._focusParentNode = oncePerAnimationFrame(this._focusParentNode).bind(this);
-    this._focusFirstNode = oncePerAnimationFrame(this._focusFirstNode).bind(this);
+    this._focusParentNode = oncePerAnimationFrame(this._focusParentNode).bind(
+      this
+    );
+    this._focusFirstNode = oncePerAnimationFrame(this._focusFirstNode).bind(
+      this
+    );
     this._focusLastNode = oncePerAnimationFrame(this._focusLastNode).bind(this);
-    this._activateNode = oncePerAnimationFrame(this._activateNode).bind(this);
 
     this._autoExpand = this._autoExpand.bind(this);
     this._preventArrowKeyScrolling = this._preventArrowKeyScrolling.bind(this);
     this._updateHeight = this._updateHeight.bind(this);
+    this._onResize = this._onResize.bind(this);
     this._dfs = this._dfs.bind(this);
     this._dfsFromRoots = this._dfsFromRoots.bind(this);
     this._focus = this._focus.bind(this);
-    this._onBlur = this._onBlur.bind(this);
+    this._activate = this._activate.bind(this);
     this._onKeyDown = this._onKeyDown.bind(this);
   }
 
   componentDidMount() {
-    window.addEventListener("resize", this._updateHeight);
+    window.addEventListener("resize", this._onResize);
     this._autoExpand();
     this._updateHeight();
+    this._scrollItemIntoView();
   }
 
   componentWillReceiveProps(nextProps) {
@@ -273,16 +309,31 @@ class Tree extends Component {
   }
 
   shouldComponentUpdate(nextProps, nextState) {
-    let { scroll, height, seen, mouseDown } = this.state;
+    const { scroll, height, seen, mouseDown } = this.state;
 
-    return scroll !== nextState.scroll ||
-           height !== nextState.height ||
-           seen !== nextState.seen ||
-           mouseDown === nextState.mouseDown;
+    return (
+      scroll !== nextState.scroll ||
+      height !== nextState.height ||
+      seen !== nextState.seen ||
+      mouseDown === nextState.mouseDown
+    );
+  }
+
+  componentDidUpdate() {
+    this._scrollItemIntoView();
   }
 
   componentWillUnmount() {
-    window.removeEventListener("resize", this._updateHeight);
+    window.removeEventListener("resize", this._onResize);
+  }
+
+  _scrollItemIntoView() {
+    const { shown } = this.props;
+    if (!shown) {
+      return;
+    }
+
+    this._scrollIntoView(shown);
   }
 
   _autoExpand() {
@@ -294,8 +345,10 @@ class Tree extends Component {
     // not use the usual DFS infrastructure because we don't want to ignore
     // collapsed nodes.
     const autoExpand = (item, currentDepth) => {
-      if (currentDepth >= this.props.autoExpandDepth ||
-          this.state.seen.has(item)) {
+      if (
+        currentDepth >= this.props.autoExpandDepth ||
+        this.state.seen.has(item)
+      ) {
         return;
       }
 
@@ -322,16 +375,8 @@ class Tree extends Component {
       case "ArrowDown":
       case "ArrowLeft":
       case "ArrowRight":
-        e.preventDefault();
-        e.stopPropagation();
-        if (e.nativeEvent) {
-          if (e.nativeEvent.preventDefault) {
-            e.nativeEvent.preventDefault();
-          }
-          if (e.nativeEvent.stopPropagation) {
-            e.nativeEvent.stopPropagation();
-          }
-        }
+        preventDefaultAndStopPropagation(e);
+        break;
     }
   }
 
@@ -413,6 +458,60 @@ class Tree extends Component {
     }
   }
 
+  /**
+   * Scroll item into view. Depending on whether the item is already rendered,
+   * we might have to calculate the position of the item based on its index and
+   * the item height.
+   *
+   * @param {Object} item
+   *        The item to be scrolled into view.
+   * @param {Number|undefined} index
+   *        The index of the item in a full DFS traversal (ignoring collapsed
+   *        nodes) or undefined.
+   * @param {Object} options
+   *        Optional information regarding item's requested alignement when
+   *        scrolling.
+   */
+  _scrollIntoView(item, index, options = {}) {
+    const treeElement = this.refs.tree;
+    if (!treeElement) {
+      return;
+    }
+
+    const element = document.getElementById(this.props.getKey(item));
+    if (element) {
+      scrollIntoView(element, { ...options, container: treeElement });
+      return;
+    }
+
+    if (index == null) {
+      // If index is not provided, determine item index from traversal.
+      const traversal = this._dfsFromRoots();
+      index = traversal.findIndex(({ item: i }) => i === item);
+    }
+
+    if (index == null || index < 0) {
+      return;
+    }
+
+    const { itemHeight } = this.props;
+    const { clientHeight, scrollTop } = treeElement;
+    const elementTop = index * itemHeight;
+    let scrollTo;
+    if (scrollTop >= elementTop + itemHeight) {
+      scrollTo = elementTop;
+    } else if (scrollTop + clientHeight <= elementTop) {
+      scrollTo = elementTop + itemHeight - clientHeight;
+    }
+
+    if (scrollTo != undefined) {
+      treeElement.scrollTo({
+        left: 0,
+        top: scrollTo,
+      });
+    }
+  }
+
   /**
    * Sets the passed in item to be the focused item.
    *
@@ -423,21 +522,15 @@ class Tree extends Component {
    * @param {Object|undefined} item
    *        The item to be focused, or undefined to focus no item.
    */
-  _focus(index, item) {
-    if (item !== undefined) {
-      const itemStartPosition = index * this.props.itemHeight;
-      const itemEndPosition = (index + 1) * this.props.itemHeight;
+  _focus(index, item, options = {}) {
+    if (item !== undefined && !options.preventAutoScroll) {
+      this._scrollIntoView(item, index, options);
+    }
 
-      // Note that if the height of the viewport (this.state.height) is less
-      // than `this.props.itemHeight`, we could accidentally try and scroll both
-      // up and down in a futile attempt to make both the item's start and end
-      // positions visible. Instead, give priority to the start of the item by
-      // checking its position first, and then using an "else if", rather than
-      // a separate "if", for the end position.
-      if (this.state.scroll > itemStartPosition) {
-        this.refs.tree.scrollTo(0, itemStartPosition);
-      } else if ((this.state.scroll + this.state.height) < itemEndPosition) {
-        this.refs.tree.scrollTo(0, itemEndPosition - this.state.height);
+    if (this.props.active != null) {
+      this._activate(null);
+      if (this.refs.tree !== this.activeElement) {
+        this.refs.tree.focus();
       }
     }
 
@@ -446,11 +539,24 @@ class Tree extends Component {
     }
   }
 
+  _activate(item) {
+    if (this.props.onActivate) {
+      this.props.onActivate(item);
+    }
+  }
+
   /**
-   * Sets the state to have no focused item.
+   * Update state height and tree's scrollTop if necessary.
    */
-  _onBlur() {
-    this._focus(0, undefined);
+  _onResize() {
+    // When tree size changes without direct user action, scroll top cat get re-set to 0
+    // (for example, when tree height changes via CSS rule change). We need to ensure that
+    // the tree's scrollTop is in sync with the scroll state.
+    if (this.state.scroll !== this.refs.tree.scrollTop) {
+      this.refs.tree.scrollTo({ left: 0, top: this.state.scroll });
+    }
+
+    this._updateHeight();
   }
 
   /**
@@ -462,7 +568,7 @@ class Tree extends Component {
   _onScroll(e) {
     this.setState({
       scroll: Math.max(this.refs.tree.scrollTop, 0),
-      height: this.refs.tree.clientHeight
+      height: this.refs.tree.clientHeight,
     });
   }
 
@@ -471,6 +577,7 @@ class Tree extends Component {
    *
    * @param {Event} e
    */
+  // eslint-disable-next-line complexity
   _onKeyDown(e) {
     if (this.props.focused == null) {
       return;
@@ -493,8 +600,10 @@ class Tree extends Component {
         break;
 
       case "ArrowLeft":
-        if (this.props.isExpanded(this.props.focused)
-            && this.props.getChildren(this.props.focused).length) {
+        if (
+          this.props.isExpanded(this.props.focused) &&
+          this.props.getChildren(this.props.focused).length
+        ) {
           this._onCollapse(this.props.focused);
         } else {
           this._focusParentNode();
@@ -502,9 +611,12 @@ class Tree extends Component {
         break;
 
       case "ArrowRight":
-        if (!this.props.isExpanded(this.props.focused)) {
+        if (
+          this.props.getChildren(this.props.focused).length &&
+          !this.props.isExpanded(this.props.focused)
+        ) {
           this._onExpand(this.props.focused);
-        } else {
+        } else if (!this.props.preventNavigationOnArrowRight) {
           this._focusNextNode();
         }
         break;
@@ -519,26 +631,42 @@ class Tree extends Component {
 
       case "Enter":
       case " ":
-        this._activateNode();
+        // On space or enter make focused tree node active. This means keyboard focus
+        // handling is passed on to the tree node itself.
+        if (this.refs.tree === this.activeElement) {
+          preventDefaultAndStopPropagation(e);
+          if (this.props.active !== this.props.focused) {
+            this._activate(this.props.focused);
+          }
+        }
+        break;
+
+      case "Escape":
+        preventDefaultAndStopPropagation(e);
+        if (this.props.active != null) {
+          this._activate(null);
+        }
+
+        if (this.refs.tree !== this.activeElement) {
+          this.refs.tree.focus();
+        }
         break;
     }
   }
 
-  _activateNode() {
-    if (this.props.onActivate) {
-      this.props.onActivate(this.props.focused);
-    }
+  get activeElement() {
+    return this.refs.tree.ownerDocument.activeElement;
   }
 
   _focusFirstNode() {
     const traversal = this._dfsFromRoots();
-    this._focus(0, traversal[0].item);
+    this._focus(0, traversal[0].item, { alignTo: "top" });
   }
 
   _focusLastNode() {
     const traversal = this._dfsFromRoots();
     const lastIndex = traversal.length - 1;
-    this._focus(lastIndex, traversal[lastIndex].item);
+    this._focus(lastIndex, traversal[lastIndex].item, { alignTo: "bottom" });
   }
 
   /**
@@ -567,7 +695,7 @@ class Tree extends Component {
       return;
     }
 
-    this._focus(prevIndex, prev);
+    this._focus(prevIndex, prev, { alignTo: "top" });
   }
 
   /**
@@ -591,7 +719,7 @@ class Tree extends Component {
     }
 
     if (i + 1 < traversal.length) {
-      this._focus(i + 1, traversal[i + 1].item);
+      this._focus(i + 1, traversal[i + 1].item, { alignTo: "bottom" });
     }
   }
 
@@ -614,7 +742,7 @@ class Tree extends Component {
       }
     }
 
-    this._focus(parentIndex, parent);
+    this._focus(parentIndex, parent, { alignTo: "top" });
   }
 
   render() {
@@ -626,10 +754,14 @@ class Tree extends Component {
     // the top and bottom of the page are filled with the `NUMBER_OF_OFFSCREEN_ITEMS`
     // previous and next items respectively, which helps the user to see fewer empty
     // gaps when scrolling quickly.
-    const { itemHeight, focused } = this.props;
+    const { itemHeight, active, focused } = this.props;
     const { scroll, height } = this.state;
-    const begin = Math.max(((scroll / itemHeight) | 0) - NUMBER_OF_OFFSCREEN_ITEMS, 0);
-    const end = Math.ceil((scroll + height) / itemHeight) + NUMBER_OF_OFFSCREEN_ITEMS;
+    const begin = Math.max(
+      ((scroll / itemHeight) | 0) - NUMBER_OF_OFFSCREEN_ITEMS,
+      0
+    );
+    const end =
+      Math.ceil((scroll + height) / itemHeight) + NUMBER_OF_OFFSCREEN_ITEMS;
     const toRender = traversal.slice(begin, end);
     const topSpacerHeight = begin * itemHeight;
     const bottomSpacerHeight = Math.max(traversal.length - end, 0) * itemHeight;
@@ -641,9 +773,9 @@ class Tree extends Component {
         style: {
           padding: 0,
           margin: 0,
-          height: topSpacerHeight + "px"
-        }
-      })
+          height: topSpacerHeight + "px",
+        },
+      }),
     ];
 
     for (let i = 0; i < toRender.length; i++) {
@@ -652,33 +784,44 @@ class Tree extends Component {
       const last = index == traversal.length - 1;
       const { item, depth } = toRender[i];
       const key = this.props.getKey(item);
-      nodes.push(TreeNode({
-        key,
-        index,
-        first,
-        last,
-        item,
-        depth,
-        id: key,
-        renderItem: this.props.renderItem,
-        focused: focused === item,
-        expanded: this.props.isExpanded(item),
-        hasChildren: !!this.props.getChildren(item).length,
-        onExpand: this._onExpand,
-        onCollapse: this._onCollapse,
-        onClick: () => this._focus(begin + i, item),
-      }));
+      nodes.push(
+        TreeNode({
+          // We make a key unique depending on whether the tree node is in active or
+          // inactive state to make sure that it is actually replaced and the tabbable
+          // state is reset.
+          key: `${key}-${active === item ? "active" : "inactive"}`,
+          index,
+          first,
+          last,
+          item,
+          depth,
+          id: key,
+          renderItem: this.props.renderItem,
+          focused: focused === item,
+          active: active === item,
+          expanded: this.props.isExpanded(item),
+          hasChildren: !!this.props.getChildren(item).length,
+          onExpand: this._onExpand,
+          onCollapse: this._onCollapse,
+          // Since the user just clicked the node, there's no need to check if
+          // it should be scrolled into view.
+          onClick: () =>
+            this._focus(begin + i, item, { preventAutoScroll: true }),
+        })
+      );
     }
 
-    nodes.push(dom.div({
-      key: "bottom-spacer",
-      role: "presentation",
-      style: {
-        padding: 0,
-        margin: 0,
-        height: bottomSpacerHeight + "px"
-      }
-    }));
+    nodes.push(
+      dom.div({
+        key: "bottom-spacer",
+        role: "presentation",
+        style: {
+          padding: 0,
+          margin: 0,
+          height: bottomSpacerHeight + "px",
+        },
+      })
+    );
 
     return dom.div(
       {
@@ -702,6 +845,14 @@ class Tree extends Component {
           // interarction.
           this._focus(begin, toRender[0].item);
         },
+        onBlur: e => {
+          if (active != null) {
+            const { relatedTarget } = e;
+            if (!this.refs.tree.contains(relatedTarget)) {
+              this._activate(null);
+            }
+          }
+        },
         onClick: () => {
           // Focus should always remain on the tree container itself.
           this.refs.tree.focus();
@@ -711,8 +862,8 @@ class Tree extends Component {
         "aria-activedescendant": focused && this.props.getKey(focused),
         style: {
           padding: 0,
-          margin: 0
-        }
+          margin: 0,
+        },
       },
       nodes
     );
@@ -735,17 +886,22 @@ class ArrowExpanderClass extends Component {
   }
 
   shouldComponentUpdate(nextProps, nextState) {
-    return this.props.item !== nextProps.item
-      || this.props.visible !== nextProps.visible
-      || this.props.expanded !== nextProps.expanded;
+    return (
+      this.props.item !== nextProps.item ||
+      this.props.visible !== nextProps.visible ||
+      this.props.expanded !== nextProps.expanded
+    );
   }
 
   render() {
     const attrs = {
       className: "arrow theme-twisty",
+      // To collapse/expand the tree rows use left/right arrow keys.
+      tabIndex: "-1",
+      "aria-hidden": true,
       onClick: this.props.expanded
         ? () => this.props.onCollapse(this.props.item)
-        : e => this.props.onExpand(this.props.item, e.altKey)
+        : e => this.props.onExpand(this.props.item, e.altKey),
     };
 
     if (this.props.expanded) {
@@ -754,7 +910,7 @@ class ArrowExpanderClass extends Component {
 
     if (!this.props.visible) {
       attrs.style = {
-        visibility: "hidden"
+        visibility: "hidden",
       };
     }
 
@@ -767,6 +923,7 @@ class TreeNodeClass extends Component {
     return {
       id: PropTypes.any.isRequired,
       focused: PropTypes.bool.isRequired,
+      active: PropTypes.bool.isRequired,
       item: PropTypes.any.isRequired,
       expanded: PropTypes.bool.isRequired,
       hasChildren: PropTypes.bool.isRequired,
@@ -781,6 +938,52 @@ class TreeNodeClass extends Component {
     };
   }
 
+  constructor(props) {
+    super(props);
+
+    this._onKeyDown = this._onKeyDown.bind(this);
+  }
+
+  componentDidMount() {
+    // Make sure that none of the focusable elements inside the tree node container are
+    // tabbable if the tree node is not active. If the tree node is active and focus is
+    // outside its container, focus on the first focusable element inside.
+    const elms = getFocusableElements(this.refs.treenode);
+    if (elms.length === 0) {
+      return;
+    }
+
+    if (!this.props.active) {
+      elms.forEach(elm => elm.setAttribute("tabindex", "-1"));
+      return;
+    }
+
+    if (!elms.includes(this.refs.treenode.ownerDocument.activeElement)) {
+      elms[0].focus();
+    }
+  }
+
+  _onKeyDown(e) {
+    const { target, key, shiftKey } = e;
+
+    if (key !== "Tab") {
+      return;
+    }
+
+    const focusMoved = !!wrapMoveFocus(
+      getFocusableElements(this.refs.treenode),
+      target,
+      shiftKey
+    );
+    if (focusMoved) {
+      // Focus was moved to the begining/end of the list, so we need to prevent the
+      // default focus change that would happen here.
+      e.preventDefault();
+    }
+
+    e.stopPropagation();
+  }
+
   render() {
     const arrow = ArrowExpander({
       item: this.props.item,
@@ -790,7 +993,7 @@ class TreeNodeClass extends Component {
       onCollapse: this.props.onCollapse,
     });
 
-    let classList = [ "tree-node", "div" ];
+    const classList = ["tree-node", "div"];
     if (this.props.index % 2) {
       classList.push("tree-node-odd");
     }
@@ -800,6 +1003,9 @@ class TreeNodeClass extends Component {
     if (this.props.last) {
       classList.push("tree-node-last");
     }
+    if (this.props.active) {
+      classList.push("tree-node-active");
+    }
 
     let ariaExpanded;
     if (this.props.hasChildren) {
@@ -814,22 +1020,26 @@ class TreeNodeClass extends Component {
         id: this.props.id,
         className: classList.join(" "),
         role: "treeitem",
+        ref: "treenode",
         "aria-level": this.props.depth + 1,
         onClick: this.props.onClick,
+        onKeyDownCapture: this.props.active ? this._onKeyDown : undefined,
         "aria-expanded": ariaExpanded,
         "data-expanded": this.props.expanded ? "" : undefined,
         "data-depth": this.props.depth,
         style: {
           padding: 0,
-          margin: 0
-        }
+          margin: 0,
+        },
       },
 
-      this.props.renderItem(this.props.item,
-                            this.props.depth,
-                            this.props.focused,
-                            arrow,
-                            this.props.expanded),
+      this.props.renderItem(
+        this.props.item,
+        this.props.depth,
+        this.props.focused,
+        arrow,
+        this.props.expanded
+      )
     );
   }
 }
@@ -847,7 +1057,7 @@ const TreeNode = createFactory(TreeNodeClass);
 function oncePerAnimationFrame(fn) {
   let animationId = null;
   let argsToPass = null;
-  return function (...args) {
+  return function(...args) {
     argsToPass = args;
     if (animationId !== null) {
       return;
