@@ -1,5 +1,3 @@
-/* -*- indent-tabs-mode: nil; js-indent-level: 2 -*- */
-/* vim: set ft=javascript ts=2 et sw=2 tw=80: */
 /* Any copyright is dedicated to the Public Domain.
  * http://creativecommons.org/publicdomain/zero/1.0/ */
 
@@ -24,12 +22,9 @@ add_task(async function testNormalExecution() {
     WORKER_URL
   );
 
-  const jsterm = await getSplitConsole(toolbox);
-  const executed = await jsterm.execute("this.location.toString()");
-  ok(
-    executed.textContent.includes(WORKER_URL),
-    "Evaluating the global's location works"
-  );
+  const hud = await getSplitConsole(toolbox);
+  await executeAndWaitForMessage(hud, "this.location.toString()", WORKER_URL);
+  ok(true, "Evaluating the global's location works");
 
   terminateWorkerInTab(tab, WORKER_URL);
   await waitForWorkerClose(workerTargetFront);
