@@ -9,10 +9,10 @@
 // Test that cookies with domain equal to full host name are listed.
 // E.g., ".example.org" vs. example.org). Bug 1149497.
 
-add_task(function* () {
-  yield openTabAndSetupStorage(MAIN_DOMAIN + "storage-cookies.html");
+add_task(async function() {
+  await openTabAndSetupStorage(MAIN_DOMAIN + "storage-cookies.html");
 
-  yield checkState([
+  await checkState([
     [
       ["cookies", "http://test1.example.org"],
       [
@@ -20,10 +20,10 @@ add_task(function* () {
         getCookieId("test2", "test1.example.org", "/browser"),
         getCookieId("test3", ".test1.example.org", "/browser"),
         getCookieId("test4", "test1.example.org", "/browser"),
-        getCookieId("test5", ".test1.example.org", "/browser")
-      ]
+        getCookieId("test5", ".test1.example.org", "/browser"),
+      ],
     ],
   ]);
 
-  yield finishTests();
+  await finishTests();
 });
