@@ -1,5 +1,5 @@
-/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
- * vim: set ts=8 sts=4 et sw=4 tw=99:
+/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
+ * vim: set ts=8 sts=2 et sw=2 tw=80:
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -9,8 +9,13 @@
 
 #include "vm/NativeObject.h"
 
+#include "mozilla/Maybe.h"
+
 #include "builtin/TypedObject.h"
-#include "gc/GCTrace.h"
+#include "gc/Allocator.h"
+#include "gc/GCProbes.h"
+#include "gc/MaybeRooted.h"
+#include "js/Result.h"
 #include "proxy/Proxy.h"
 #include "vm/JSContext.h"
 #include "vm/ProxyObject.h"
@@ -23,15 +28,19 @@
 
 namespace js {
 
+inline uint32_t NativeObject::numFixedSlotsMaybeForwarded() const {
+  return gc::MaybeForwarded(lastProperty())->numFixedSlots();
+}
+
 inline uint8_t* NativeObject::fixedData(size_t nslots) const {
   MOZ_ASSERT(ClassCanHaveFixedData(getClass()));
-  MOZ_ASSERT(nslots == numFixedSlots() + (hasPrivate() ? 1 : 0));
+  MOZ_ASSERT(nslots == numFixedSlotsMaybeForwarded() + (hasPrivate() ? 1 : 0));
   return reinterpret_cast<uint8_t*>(&fixedSlots()[nslots]);
 }
 
 inline void NativeObject::removeLastProperty(JSContext* cx) {
   MOZ_ASSERT(canRemoveLastProperty());
-  JS_ALWAYS_TRUE(setLastProperty(cx, lastProperty()->previous()));
+  MOZ_ALWAYS_TRUE(setLastProperty(cx, lastProperty()->previous()));
 }
 
 inline bool NativeObject::canRemoveLastProperty() {
@@ -57,13 +66,24 @@ inline void NativeObject::clearShouldConvertDoubleElements() {
   getElementsHeader()->clearShouldConvertDoubleElements();
 }
 
-inline void NativeObject::setDenseElementWithType(JSContext* cx, uint32_t index,
-                                                  const Value& val) {
+inline void NativeObject::addDenseElementType(JSContext* cx, uint32_t index,
+                                              const Value& val) {
+  if (!IsTypeInferenceEnabled()) {
+    return;
+  }
+
   // Avoid a slow AddTypePropertyId call if the type is the same as the type
   // of the previous element.
   TypeSet::Type thisType = TypeSet::GetValueType(val);
-  if (index == 0 || TypeSet::GetValueType(elements_[index - 1]) != thisType)
+  if (index == 0 || elements_[index - 1].isMagic() ||
+      TypeSet::GetValueType(elements_[index - 1]) != thisType) {
     AddTypePropertyId(cx, this, JSID_VOID, thisType);
+  }
+}
+
+inline void NativeObject::setDenseElementWithType(JSContext* cx, uint32_t index,
+                                                  const Value& val) {
+  addDenseElementType(cx, index, val);
   setDenseElementMaybeConvertDouble(index, val);
 }
 
@@ -71,24 +91,29 @@ inline void NativeObject::initDenseElementWithType(JSContext* cx,
                                                    uint32_t index,
                                                    const Value& val) {
   MOZ_ASSERT(!shouldConvertDoubleElements());
-  if (val.isMagic(JS_ELEMENTS_HOLE))
-    markDenseElementsNotPacked(cx);
-  else
-    AddTypePropertyId(cx, this, JSID_VOID, val);
+  MOZ_ASSERT(!val.isMagic(JS_ELEMENTS_HOLE));
+
+  addDenseElementType(cx, index, val);
   initDenseElement(index, val);
 }
 
 inline void NativeObject::setDenseElementHole(JSContext* cx, uint32_t index) {
-  MarkObjectGroupFlags(cx, this, OBJECT_FLAG_NON_PACKED);
+  if (IsTypeInferenceEnabled()) {
+    MarkObjectGroupFlags(cx, this, OBJECT_FLAG_NON_PACKED);
+  }
   setDenseElement(index, MagicValue(JS_ELEMENTS_HOLE));
 }
 
-/* static */ inline void NativeObject::removeDenseElementForSparseIndex(
-    JSContext* cx, HandleNativeObject obj, uint32_t index) {
-  MarkObjectGroupFlags(cx, obj,
-                       OBJECT_FLAG_NON_PACKED | OBJECT_FLAG_SPARSE_INDEXES);
-  if (obj->containsDenseElement(index))
-    obj->setDenseElementUnchecked(index, MagicValue(JS_ELEMENTS_HOLE));
+inline void NativeObject::removeDenseElementForSparseIndex(JSContext* cx,
+                                                           uint32_t index) {
+  MOZ_ASSERT(containsPure(INT_TO_JSID(index)));
+  if (IsTypeInferenceEnabled()) {
+    MarkObjectGroupFlags(cx, this,
+                         OBJECT_FLAG_NON_PACKED | OBJECT_FLAG_SPARSE_INDEXES);
+  }
+  if (containsDenseElement(index)) {
+    setDenseElement(index, MagicValue(JS_ELEMENTS_HOLE));
+  }
 }
 
 inline bool NativeObject::writeToIndexWouldMarkNotPacked(uint32_t index) {
@@ -97,17 +122,24 @@ inline bool NativeObject::writeToIndexWouldMarkNotPacked(uint32_t index) {
 
 inline void NativeObject::markDenseElementsNotPacked(JSContext* cx) {
   MOZ_ASSERT(isNative());
-  MarkObjectGroupFlags(cx, this, OBJECT_FLAG_NON_PACKED);
+  if (IsTypeInferenceEnabled()) {
+    MarkObjectGroupFlags(cx, this, OBJECT_FLAG_NON_PACKED);
+  }
 }
 
 inline void NativeObject::elementsRangeWriteBarrierPost(uint32_t start,
                                                         uint32_t count) {
+  if (!isTenured()) {
+    return;
+  }
   for (size_t i = 0; i < count; i++) {
     const Value& v = elements_[start + i];
-    if ((v.isObject() || v.isString()) && IsInsideNursery(v.toGCThing())) {
-      zone()->group()->storeBuffer().putSlot(
-          this, HeapSlot::Element, unshiftedIndex(start + i), count - i);
-      return;
+    if (v.isGCThing()) {
+      if (gc::StoreBuffer* sb = v.toGCThing()->storeBuffer()) {
+        sb->putSlot(this, HeapSlot::Element, unshiftedIndex(start + i),
+                    count - i);
+        return;
+      }
     }
   }
 }
@@ -116,13 +148,17 @@ inline void NativeObject::copyDenseElements(uint32_t dstStart, const Value* src,
                                             uint32_t count) {
   MOZ_ASSERT(dstStart + count <= getDenseCapacity());
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible());
   MOZ_ASSERT_IF(count > 0, src != nullptr);
 #ifdef DEBUG
-  for (uint32_t i = 0; i < count; ++i) checkStoredValue(src[i]);
+  for (uint32_t i = 0; i < count; ++i) {
+    checkStoredValue(src[i]);
+  }
 #endif
-  if (count == 0) return;
-  if (JS::shadow::Zone::asShadowZone(zone())->needsIncrementalBarrier()) {
+  if (count == 0) {
+    return;
+  }
+  if (zone()->needsIncrementalBarrier()) {
     uint32_t numShifted = getElementsHeader()->numShiftedElements();
     for (uint32_t i = 0; i < count; ++i) {
       elements_[dstStart + i].set(this, HeapSlot::Element,
@@ -147,12 +183,14 @@ inline void NativeObject::initDenseElements(const Value* src, uint32_t count) {
   MOZ_ASSERT(getDenseInitializedLength() == 0);
   MOZ_ASSERT(count <= getDenseCapacity());
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible());
 
   setDenseInitializedLength(count);
 
 #ifdef DEBUG
-  for (uint32_t i = 0; i < count; ++i) checkStoredValue(src[i]);
+  for (uint32_t i = 0; i < count; ++i) {
+    checkStoredValue(src[i]);
+  }
 #endif
 
   memcpy(reinterpret_cast<Value*>(elements_), src, count * sizeof(Value));
@@ -160,10 +198,12 @@ inline void NativeObject::initDenseElements(const Value* src, uint32_t count) {
 }
 
 inline bool NativeObject::tryShiftDenseElements(uint32_t count) {
+  MOZ_ASSERT(isExtensible());
+
   ObjectElements* header = getElementsHeader();
   if (header->initializedLength == count ||
       count > ObjectElements::MaxShiftedElements || header->isCopyOnWrite() ||
-      header->isFrozen() || header->hasNonwritableArrayLength()) {
+      header->hasNonwritableArrayLength()) {
     return false;
   }
 
@@ -172,6 +212,8 @@ inline bool NativeObject::tryShiftDenseElements(uint32_t count) {
 }
 
 inline void NativeObject::shiftDenseElementsUnchecked(uint32_t count) {
+  MOZ_ASSERT(isExtensible());
+
   ObjectElements* header = getElementsHeader();
   MOZ_ASSERT(count > 0);
   MOZ_ASSERT(count < header->initializedLength);
@@ -195,7 +237,7 @@ inline void NativeObject::moveDenseElements(uint32_t dstStart,
   MOZ_ASSERT(dstStart + count <= getDenseCapacity());
   MOZ_ASSERT(srcStart + count <= getDenseInitializedLength());
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible());
 
   /*
    * Using memmove here would skip write barriers. Also, we need to consider
@@ -209,18 +251,20 @@ inline void NativeObject::moveDenseElements(uint32_t dstStart,
    * write barrier is invoked here on B, despite the fact that it exists in
    * the array before and after the move.
    */
-  if (JS::shadow::Zone::asShadowZone(zone())->needsIncrementalBarrier()) {
+  if (zone()->needsIncrementalBarrier()) {
     uint32_t numShifted = getElementsHeader()->numShiftedElements();
     if (dstStart < srcStart) {
       HeapSlot* dst = elements_ + dstStart;
       HeapSlot* src = elements_ + srcStart;
-      for (uint32_t i = 0; i < count; i++, dst++, src++)
+      for (uint32_t i = 0; i < count; i++, dst++, src++) {
         dst->set(this, HeapSlot::Element, dst - elements_ + numShifted, *src);
+      }
     } else {
       HeapSlot* dst = elements_ + dstStart + count - 1;
       HeapSlot* src = elements_ + srcStart + count - 1;
-      for (uint32_t i = 0; i < count; i++, dst--, src--)
+      for (uint32_t i = 0; i < count; i++, dst--, src--) {
         dst->set(this, HeapSlot::Element, dst - elements_ + numShifted, *src);
+      }
     }
   } else {
     memmove(elements_ + dstStart, elements_ + srcStart,
@@ -232,22 +276,22 @@ inline void NativeObject::moveDenseElements(uint32_t dstStart,
 inline void NativeObject::moveDenseElementsNoPreBarrier(uint32_t dstStart,
                                                         uint32_t srcStart,
                                                         uint32_t count) {
-  MOZ_ASSERT(!shadowZone()->needsIncrementalBarrier());
+  MOZ_ASSERT(!zone()->needsIncrementalBarrier());
 
   MOZ_ASSERT(dstStart + count <= getDenseCapacity());
   MOZ_ASSERT(srcStart + count <= getDenseCapacity());
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible());
 
   memmove(elements_ + dstStart, elements_ + srcStart, count * sizeof(HeapSlot));
   elementsRangeWriteBarrierPost(dstStart, count);
 }
 
 inline void NativeObject::reverseDenseElementsNoPreBarrier(uint32_t length) {
-  MOZ_ASSERT(!shadowZone()->needsIncrementalBarrier());
+  MOZ_ASSERT(!zone()->needsIncrementalBarrier());
 
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible());
 
   MOZ_ASSERT(length > 1);
   MOZ_ASSERT(length <= getDenseInitializedLength());
@@ -271,6 +315,7 @@ inline void NativeObject::ensureDenseInitializedLengthNoPackedCheck(
     uint32_t index, uint32_t extra) {
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
   MOZ_ASSERT(!denseElementsAreFrozen());
+  MOZ_ASSERT(isExtensible() || (containsDenseElement(index) && extra == 1));
 
   /*
    * Ensure that the array's contents have been initialized up to index, and
@@ -281,9 +326,10 @@ inline void NativeObject::ensureDenseInitializedLengthNoPackedCheck(
   uint32_t& initlen = getElementsHeader()->initializedLength;
 
   if (initlen < index + extra) {
+    MOZ_ASSERT(isExtensible());
     uint32_t numShifted = getElementsHeader()->numShiftedElements();
     size_t offset = initlen;
-    for (HeapSlot *sp = elements_ + initlen; sp != elements_ + (index + extra);
+    for (HeapSlot* sp = elements_ + initlen; sp != elements_ + (index + extra);
          sp++, offset++) {
       sp->init(this, HeapSlot::Element, offset + numShifted,
                MagicValue(JS_ELEMENTS_HOLE));
@@ -295,7 +341,11 @@ inline void NativeObject::ensureDenseInitializedLengthNoPackedCheck(
 inline void NativeObject::ensureDenseInitializedLength(JSContext* cx,
                                                        uint32_t index,
                                                        uint32_t extra) {
-  if (writeToIndexWouldMarkNotPacked(index)) markDenseElementsNotPacked(cx);
+  MOZ_ASSERT(isExtensible());
+
+  if (writeToIndexWouldMarkNotPacked(index)) {
+    markDenseElementsNotPacked(cx);
+  }
   ensureDenseInitializedLengthNoPackedCheck(index, extra);
 }
 
@@ -303,24 +353,16 @@ DenseElementResult NativeObject::extendDenseElements(JSContext* cx,
                                                      uint32_t requiredCapacity,
                                                      uint32_t extra) {
   MOZ_ASSERT(!denseElementsAreCopyOnWrite());
-  MOZ_ASSERT(!denseElementsAreFrozen());
-
-  /*
-   * Don't grow elements for non-extensible objects. Dense elements can be
-   * added/written with no extensible checks as long as there is capacity
-   * for them.
-   */
-  if (!nonProxyIsExtensible()) {
-    MOZ_ASSERT(getDenseCapacity() == 0);
-    return DenseElementResult::Incomplete;
-  }
+  MOZ_ASSERT(isExtensible());
 
   /*
    * Don't grow elements for objects which already have sparse indexes.
    * This avoids needing to count non-hole elements in willBeSparseElements
    * every time a new index is added.
    */
-  if (isIndexed()) return DenseElementResult::Incomplete;
+  if (isIndexed()) {
+    return DenseElementResult::Incomplete;
+  }
 
   /*
    * We use the extra argument also as a hint about number of non-hole
@@ -331,7 +373,9 @@ DenseElementResult NativeObject::extendDenseElements(JSContext* cx,
     return DenseElementResult::Incomplete;
   }
 
-  if (!growElements(cx, requiredCapacity)) return DenseElementResult::Failure;
+  if (!growElements(cx, requiredCapacity)) {
+    return DenseElementResult::Failure;
+  }
 
   return DenseElementResult::Success;
 }
@@ -340,10 +384,15 @@ inline DenseElementResult NativeObject::ensureDenseElements(JSContext* cx,
                                                             uint32_t index,
                                                             uint32_t extra) {
   MOZ_ASSERT(isNative());
+  MOZ_ASSERT(isExtensible() || (containsDenseElement(index) && extra == 1));
 
-  if (writeToIndexWouldMarkNotPacked(index)) markDenseElementsNotPacked(cx);
+  if (writeToIndexWouldMarkNotPacked(index)) {
+    markDenseElementsNotPacked(cx);
+  }
 
-  if (!maybeCopyElementsForWrite(cx)) return DenseElementResult::Failure;
+  if (!maybeCopyElementsForWrite(cx)) {
+    return DenseElementResult::Failure;
+  }
 
   uint32_t currentCapacity = getDenseCapacity();
 
@@ -372,7 +421,9 @@ inline DenseElementResult NativeObject::ensureDenseElements(JSContext* cx,
   }
 
   DenseElementResult result = extendDenseElements(cx, requiredCapacity, extra);
-  if (result != DenseElementResult::Success) return result;
+  if (result != DenseElementResult::Success) {
+    return result;
+  }
 
   ensureDenseInitializedLengthNoPackedCheck(index, extra);
   return DenseElementResult::Success;
@@ -381,7 +432,9 @@ inline DenseElementResult NativeObject::ensureDenseElements(JSContext* cx,
 inline DenseElementResult NativeObject::setOrExtendDenseElements(
     JSContext* cx, uint32_t start, const Value* vp, uint32_t count,
     ShouldUpdateTypes updateTypes) {
-  if (denseElementsAreFrozen()) return DenseElementResult::Incomplete;
+  if (!isExtensible()) {
+    return DenseElementResult::Incomplete;
+  }
 
   if (is<ArrayObject>() && !as<ArrayObject>().lengthIsWritable() &&
       start + count >= as<ArrayObject>().length()) {
@@ -389,56 +442,35 @@ inline DenseElementResult NativeObject::setOrExtendDenseElements(
   }
 
   DenseElementResult result = ensureDenseElements(cx, start, count);
-  if (result != DenseElementResult::Success) return result;
+  if (result != DenseElementResult::Success) {
+    return result;
+  }
 
-  if (is<ArrayObject>() && start + count >= as<ArrayObject>().length())
+  if (is<ArrayObject>() && start + count >= as<ArrayObject>().length()) {
     as<ArrayObject>().setLengthInt32(start + count);
+  }
 
   if (updateTypes == ShouldUpdateTypes::DontUpdate &&
       !shouldConvertDoubleElements()) {
     copyDenseElements(start, vp, count);
   } else {
-    for (size_t i = 0; i < count; i++)
+    for (size_t i = 0; i < count; i++) {
       setDenseElementWithType(cx, start + i, vp[i]);
+    }
   }
 
   return DenseElementResult::Success;
 }
 
-inline Value NativeObject::getDenseOrTypedArrayElement(uint32_t idx) {
-  if (is<TypedArrayObject>()) return as<TypedArrayObject>().getElement(idx);
-  return getDenseElement(idx);
-}
-
-/* static */ inline NativeObject* NativeObject::copy(
-    JSContext* cx, gc::AllocKind kind, gc::InitialHeap heap,
-    HandleNativeObject templateObject) {
-  RootedShape shape(cx, templateObject->lastProperty());
-  RootedObjectGroup group(cx, templateObject->group());
-  MOZ_ASSERT(!templateObject->denseElementsAreCopyOnWrite());
-
-  JSObject* baseObj;
-  JS_TRY_VAR_OR_RETURN_NULL(cx, baseObj, create(cx, kind, heap, shape, group));
-
-  NativeObject* obj = &baseObj->as<NativeObject>();
-
-  size_t span = shape->slotSpan();
-  if (span) {
-    uint32_t numFixed = templateObject->numFixedSlots();
-    const Value* fixed = &templateObject->getSlot(0);
-    // Only copy elements which are registered in the shape, even if the
-    // number of fixed slots is larger.
-    if (span < numFixed) numFixed = span;
-    obj->copySlotRange(0, fixed, numFixed);
-
-    if (numFixed < span) {
-      uint32_t numSlots = span - numFixed;
-      const Value* slots = &templateObject->getSlot(numFixed);
-      obj->copySlotRange(numFixed, slots, numSlots);
-    }
+template <AllowGC allowGC>
+inline bool NativeObject::getDenseOrTypedArrayElement(
+    JSContext* cx, uint32_t idx,
+    typename MaybeRooted<Value, allowGC>::MutableHandleType val) {
+  if (is<TypedArrayObject>()) {
+    return as<TypedArrayObject>().getElement<allowGC>(cx, idx, val);
   }
-
-  return obj;
+  val.set(getDenseElement(idx));
+  return true;
 }
 
 MOZ_ALWAYS_INLINE void NativeObject::setSlotWithType(JSContext* cx,
@@ -447,16 +479,13 @@ MOZ_ALWAYS_INLINE void NativeObject::setSlotWithType(JSContext* cx,
                                                      bool overwriting) {
   setSlot(shape->slot(), value);
 
-  if (overwriting) shape->setOverwritten();
+  if (overwriting) {
+    shape->setOverwritten();
+  }
 
   AddTypePropertyId(cx, this, shape->propid(), value);
 }
 
-inline void NativeObject::updateShapeAfterMovingGC() {
-  Shape* shape = this->shape();
-  if (IsForwarded(shape)) shapeRef().unsafeSet(Forwarded(shape));
-}
-
 inline bool NativeObject::isInWholeCellBuffer() const {
   const gc::TenuredCell* cell = &asTenured();
   gc::ArenaCellSet* cells = cell->arena()->bufferedCells();
@@ -468,91 +497,46 @@ inline bool NativeObject::isInWholeCellBuffer() const {
     js::HandleShape shape, js::HandleObjectGroup group) {
   debugCheckNewObject(group, shape, kind, heap);
 
-  const js::Class* clasp = group->clasp();
+  const JSClass* clasp = group->clasp();
   MOZ_ASSERT(clasp->isNative());
+  MOZ_ASSERT(!clasp->isJSFunction(), "should use JSFunction::create");
 
   size_t nDynamicSlots =
       dynamicSlotsCount(shape->numFixedSlots(), shape->slotSpan(), clasp);
 
-  JSObject* obj = js::Allocate<JSObject>(cx, kind, nDynamicSlots, heap, clasp);
-  if (!obj) return cx->alreadyReportedOOM();
+  JSObject* obj = js::AllocateObject(cx, kind, nDynamicSlots, heap, clasp);
+  if (!obj) {
+    return cx->alreadyReportedOOM();
+  }
 
   NativeObject* nobj = static_cast<NativeObject*>(obj);
   nobj->initGroup(group);
   nobj->initShape(shape);
   // NOTE: Dynamic slots are created internally by Allocate<JSObject>.
-  if (!nDynamicSlots) nobj->initSlots(nullptr);
+  if (!nDynamicSlots) {
+    nobj->initSlots(nullptr);
+  }
   nobj->setEmptyElements();
 
-  if (clasp->hasPrivate()) nobj->initPrivate(nullptr);
-
-  if (size_t span = shape->slotSpan()) nobj->initializeSlotRange(0, span);
-
-  // JSFunction's fixed slots expect POD-style initialization.
-  if (clasp->isJSFunction()) {
-    MOZ_ASSERT(kind == js::gc::AllocKind::FUNCTION ||
-               kind == js::gc::AllocKind::FUNCTION_EXTENDED);
-    size_t size = kind == js::gc::AllocKind::FUNCTION
-                      ? sizeof(JSFunction)
-                      : sizeof(js::FunctionExtended);
-    memset(nobj->as<JSFunction>().fixedSlots(), 0,
-           size - sizeof(js::NativeObject));
-    if (kind == js::gc::AllocKind::FUNCTION_EXTENDED) {
-      // SetNewObjectMetadata may gc, which will be unhappy if flags &
-      // EXTENDED doesn't match the arena's AllocKind.
-      nobj->as<JSFunction>().setFlags(JSFunction::EXTENDED);
-    }
+  if (clasp->hasPrivate()) {
+    nobj->initPrivate(nullptr);
   }
 
-  if (clasp->shouldDelayMetadataBuilder())
-    cx->compartment()->setObjectPendingMetadata(cx, nobj);
-  else
-    nobj = SetNewObjectMetadata(cx, nobj);
+  if (size_t span = shape->slotSpan()) {
+    nobj->initializeSlotRange(0, span);
+  }
 
-  js::gc::TraceCreateObject(nobj);
+  if (clasp->shouldDelayMetadataBuilder()) {
+    cx->realm()->setObjectPendingMetadata(cx, nobj);
+  } else {
+    nobj = SetNewObjectMetadata(cx, nobj);
+  }
+
+  js::gc::gcprobes::CreateObject(nobj);
 
   return nobj;
 }
 
-/* static */ inline JS::Result<NativeObject*, JS::OOM&>
-NativeObject::createWithTemplate(JSContext* cx, js::gc::InitialHeap heap,
-                                 HandleObject templateObject) {
-  RootedObjectGroup group(cx, templateObject->group());
-  RootedShape shape(cx, templateObject->as<NativeObject>().lastProperty());
-
-  gc::AllocKind kind = gc::GetGCObjectKind(shape->numFixedSlots());
-  MOZ_ASSERT(CanBeFinalizedInBackground(kind, shape->getObjectClass()));
-  kind = gc::GetBackgroundAllocKind(kind);
-
-  return create(cx, kind, heap, shape, group);
-}
-
-MOZ_ALWAYS_INLINE uint32_t NativeObject::numDynamicSlots() const {
-  return dynamicSlotsCount(numFixedSlots(), slotSpan(), getClass());
-}
-
-/* static */ MOZ_ALWAYS_INLINE uint32_t NativeObject::dynamicSlotsCount(
-    uint32_t nfixed, uint32_t span, const Class* clasp) {
-  if (span <= nfixed) return 0;
-  span -= nfixed;
-
-  // Increase the slots to SLOT_CAPACITY_MIN to decrease the likelihood
-  // the dynamic slots need to get increased again. ArrayObjects ignore
-  // this because slots are uncommon in that case.
-  if (clasp != &ArrayObject::class_ && span <= SLOT_CAPACITY_MIN)
-    return SLOT_CAPACITY_MIN;
-
-  uint32_t slots = mozilla::RoundUpPow2(span);
-  MOZ_ASSERT(slots >= span);
-  return slots;
-}
-
-/* static */ MOZ_ALWAYS_INLINE uint32_t
-NativeObject::dynamicSlotsCount(Shape* shape) {
-  return dynamicSlotsCount(shape->numFixedSlots(), shape->slotSpan(),
-                           shape->getObjectClass());
-}
-
 MOZ_ALWAYS_INLINE bool NativeObject::updateSlotsForSpan(JSContext* cx,
                                                         size_t oldSpan,
                                                         size_t newSpan) {
@@ -562,18 +546,23 @@ MOZ_ALWAYS_INLINE bool NativeObject::updateSlotsForSpan(JSContext* cx,
   size_t newCount = dynamicSlotsCount(numFixedSlots(), newSpan, getClass());
 
   if (oldSpan < newSpan) {
-    if (oldCount < newCount && !growSlots(cx, oldCount, newCount)) return false;
+    if (oldCount < newCount && !growSlots(cx, oldCount, newCount)) {
+      return false;
+    }
 
-    if (newSpan == oldSpan + 1)
+    if (newSpan == oldSpan + 1) {
       initSlotUnchecked(oldSpan, UndefinedValue());
-    else
+    } else {
       initializeSlotRange(oldSpan, newSpan - oldSpan);
+    }
   } else {
     /* Trigger write barriers on the old slots before reallocating. */
     prepareSlotRangeForOverwrite(newSpan, oldSpan);
     invalidateSlotRange(newSpan, oldSpan - newSpan);
 
-    if (oldCount > newCount) shrinkSlots(cx, oldCount, newCount);
+    if (oldCount > newCount) {
+      shrinkSlots(cx, oldCount, newCount);
+    }
   }
 
   return true;
@@ -595,7 +584,9 @@ MOZ_ALWAYS_INLINE bool NativeObject::setLastProperty(JSContext* cx,
     return true;
   }
 
-  if (MOZ_UNLIKELY(!updateSlotsForSpan(cx, oldSpan, newSpan))) return false;
+  if (MOZ_UNLIKELY(!updateSlotsForSpan(cx, oldSpan, newSpan))) {
+    return false;
+  }
 
   setShape(shape);
   return true;
@@ -605,79 +596,13 @@ inline js::gc::AllocKind NativeObject::allocKindForTenure() const {
   using namespace js::gc;
   AllocKind kind = GetGCObjectFixedSlotsKind(numFixedSlots());
   MOZ_ASSERT(!IsBackgroundFinalized(kind));
-  if (!CanBeFinalizedInBackground(kind, getClass())) return kind;
-  return GetBackgroundAllocKind(kind);
+  if (!CanChangeToBackgroundAllocKind(kind, getClass())) {
+    return kind;
+  }
+  return ForegroundToBackgroundAllocKind(kind);
 }
 
-inline js::gc::AllocKind PlainObject::allocKindForTenure() const {
-  using namespace js::gc;
-  AllocKind kind = GetGCObjectFixedSlotsKind(numFixedSlots());
-  MOZ_ASSERT(!IsBackgroundFinalized(kind));
-  MOZ_ASSERT(CanBeFinalizedInBackground(kind, getClass()));
-  return GetBackgroundAllocKind(kind);
-}
-
-/* Make an object with pregenerated shape from a NEWOBJECT bytecode. */
-static inline PlainObject* CopyInitializerObject(
-    JSContext* cx, HandlePlainObject baseobj,
-    NewObjectKind newKind = GenericObject) {
-  MOZ_ASSERT(!baseobj->inDictionaryMode());
-
-  gc::AllocKind allocKind =
-      gc::GetGCObjectFixedSlotsKind(baseobj->numFixedSlots());
-  allocKind = gc::GetBackgroundAllocKind(allocKind);
-  MOZ_ASSERT_IF(baseobj->isTenured(),
-                allocKind == baseobj->asTenured().getAllocKind());
-  RootedPlainObject obj(
-      cx, NewBuiltinClassInstance<PlainObject>(cx, allocKind, newKind));
-  if (!obj) return nullptr;
-
-  if (!obj->setLastProperty(cx, baseobj->lastProperty())) return nullptr;
-
-  return obj;
-}
-
-inline NativeObject* NewNativeObjectWithGivenTaggedProto(
-    JSContext* cx, const Class* clasp, Handle<TaggedProto> proto,
-    gc::AllocKind allocKind, NewObjectKind newKind) {
-  return MaybeNativeObject(
-      NewObjectWithGivenTaggedProto(cx, clasp, proto, allocKind, newKind));
-}
-
-inline NativeObject* NewNativeObjectWithGivenTaggedProto(
-    JSContext* cx, const Class* clasp, Handle<TaggedProto> proto,
-    NewObjectKind newKind = GenericObject) {
-  return MaybeNativeObject(
-      NewObjectWithGivenTaggedProto(cx, clasp, proto, newKind));
-}
-
-inline NativeObject* NewNativeObjectWithGivenProto(JSContext* cx,
-                                                   const Class* clasp,
-                                                   HandleObject proto,
-                                                   gc::AllocKind allocKind,
-                                                   NewObjectKind newKind) {
-  return MaybeNativeObject(
-      NewObjectWithGivenProto(cx, clasp, proto, allocKind, newKind));
-}
-
-inline NativeObject* NewNativeObjectWithGivenProto(
-    JSContext* cx, const Class* clasp, HandleObject proto,
-    NewObjectKind newKind = GenericObject) {
-  return MaybeNativeObject(NewObjectWithGivenProto(cx, clasp, proto, newKind));
-}
-
-inline NativeObject* NewNativeObjectWithClassProto(
-    JSContext* cx, const Class* clasp, HandleObject proto,
-    gc::AllocKind allocKind, NewObjectKind newKind = GenericObject) {
-  return MaybeNativeObject(
-      NewObjectWithClassProto(cx, clasp, proto, allocKind, newKind));
-}
-
-inline NativeObject* NewNativeObjectWithClassProto(
-    JSContext* cx, const Class* clasp, HandleObject proto,
-    NewObjectKind newKind = GenericObject) {
-  return MaybeNativeObject(NewObjectWithClassProto(cx, clasp, proto, newKind));
-}
+inline js::GlobalObject& NativeObject::global() const { return nonCCWGlobal(); }
 
 /*
  * Call obj's resolve hook.
@@ -712,9 +637,14 @@ static MOZ_ALWAYS_INLINE bool CallResolveOp(JSContext* cx,
   *recursedp = false;
 
   bool resolved = false;
-  if (!obj->getClass()->getResolve()(cx, obj, id, &resolved)) return false;
+  AutoRealm ar(cx, obj);
+  if (!obj->getClass()->getResolve()(cx, obj, id, &resolved)) {
+    return false;
+  }
 
-  if (!resolved) return true;
+  if (!resolved) {
+    return true;
+  }
 
   // Assert the mayResolve hook, if there is one, returns true for this
   // property.
@@ -729,10 +659,11 @@ static MOZ_ALWAYS_INLINE bool CallResolveOp(JSContext* cx,
   MOZ_ASSERT(!obj->is<TypedArrayObject>());
 
   RootedShape shape(cx, obj->lookup(cx, id));
-  if (shape)
+  if (shape) {
     propp.setNativeProperty(shape);
-  else
+  } else {
     propp.setNotFound();
+  }
 
   return true;
 }
@@ -754,17 +685,28 @@ static MOZ_ALWAYS_INLINE bool LookupOwnPropertyInline(
   // so that integer properties on the prototype are ignored even for out
   // of bounds accesses.
   if (obj->template is<TypedArrayObject>()) {
-    uint64_t index;
-    if (IsTypedArrayIndex(id, &index)) {
-      if (index < obj->template as<TypedArrayObject>().length())
+    JS::Result<mozilla::Maybe<uint64_t>> index = IsTypedArrayIndex(cx, id);
+    if (index.isErr()) {
+      if (!allowGC) {
+        cx->recoverFromOutOfMemory();
+      }
+      return false;
+    }
+
+    if (index.inspect()) {
+      if (index.inspect().value() <
+          obj->template as<TypedArrayObject>().length()) {
         propp.setDenseOrTypedArrayElement();
-      else
+      } else {
         propp.setNotFound();
+      }
       *donep = true;
       return true;
     }
   }
 
+  MOZ_ASSERT(cx->compartment() == obj->compartment());
+
   // Check for a native property. Call Shape::search directly (instead of
   // NativeObject::lookup) because it's inlined.
   if (Shape* shape = obj->lastProperty()->search(cx, id)) {
@@ -775,27 +717,25 @@ static MOZ_ALWAYS_INLINE bool LookupOwnPropertyInline(
 
   // id was not found in obj. Try obj's resolve hook, if any.
   if (obj->getClass()->getResolve()) {
-    MOZ_ASSERT(!cx->helperThread());
-    if (!allowGC) return false;
-
-    bool recursed;
-    if (!CallResolveOp(
-            cx, MaybeRooted<NativeObject*, allowGC>::toHandle(obj),
-            MaybeRooted<jsid, allowGC>::toHandle(id),
-            MaybeRooted<PropertyResult, allowGC>::toMutableHandle(propp),
-            &recursed)) {
+    MOZ_ASSERT(!cx->isHelperThreadContext());
+    if constexpr (!allowGC) {
       return false;
-    }
+    } else {
+      bool recursed;
+      if (!CallResolveOp(cx, obj, id, propp, &recursed)) {
+        return false;
+      }
 
-    if (recursed) {
-      propp.setNotFound();
-      *donep = true;
-      return true;
-    }
+      if (recursed) {
+        propp.setNotFound();
+        *donep = true;
+        return true;
+      }
 
-    if (propp) {
-      *donep = true;
-      return true;
+      if (propp) {
+        *donep = true;
+        return true;
+      }
     }
   }
 
@@ -808,32 +748,37 @@ static MOZ_ALWAYS_INLINE bool LookupOwnPropertyInline(
  * Simplified version of LookupOwnPropertyInline that doesn't call resolve
  * hooks.
  */
-static inline void NativeLookupOwnPropertyNoResolve(
+static inline MOZ_MUST_USE bool NativeLookupOwnPropertyNoResolve(
     JSContext* cx, HandleNativeObject obj, HandleId id,
     MutableHandle<PropertyResult> result) {
   // Check for a native dense element.
   if (JSID_IS_INT(id) && obj->containsDenseElement(JSID_TO_INT(id))) {
     result.setDenseOrTypedArrayElement();
-    return;
+    return true;
   }
 
   // Check for a typed array element.
   if (obj->is<TypedArrayObject>()) {
-    uint64_t index;
-    if (IsTypedArrayIndex(id, &index)) {
-      if (index < obj->as<TypedArrayObject>().length())
+    mozilla::Maybe<uint64_t> index;
+    JS_TRY_VAR_OR_RETURN_FALSE(cx, index, IsTypedArrayIndex(cx, id));
+
+    if (index) {
+      if (index.value() < obj->as<TypedArrayObject>().length()) {
         result.setDenseOrTypedArrayElement();
-      else
+      } else {
         result.setNotFound();
-      return;
+      }
+      return true;
     }
   }
 
   // Check for a native property.
-  if (Shape* shape = obj->lookup(cx, id))
+  if (Shape* shape = obj->lookup(cx, id)) {
     result.setNativeProperty(shape);
-  else
+  } else {
     result.setNotFound();
+  }
+  return true;
 }
 
 template <AllowGC allowGC>
@@ -847,28 +792,31 @@ static MOZ_ALWAYS_INLINE bool LookupPropertyInline(
 
   while (true) {
     bool done;
-    if (!LookupOwnPropertyInline<allowGC>(cx, current, id, propp, &done))
+    if (!LookupOwnPropertyInline<allowGC>(cx, current, id, propp, &done)) {
       return false;
+    }
     if (done) {
-      if (propp)
+      if (propp) {
         objp.set(current);
-      else
+      } else {
         objp.set(nullptr);
+      }
       return true;
     }
 
     typename MaybeRooted<JSObject*, allowGC>::RootType proto(
         cx, current->staticPrototype());
 
-    if (!proto) break;
+    if (!proto) {
+      break;
+    }
     if (!proto->isNative()) {
-      MOZ_ASSERT(!cx->helperThread());
-      if (!allowGC) return false;
-      return LookupProperty(
-          cx, MaybeRooted<JSObject*, allowGC>::toHandle(proto),
-          MaybeRooted<jsid, allowGC>::toHandle(id),
-          MaybeRooted<JSObject*, allowGC>::toMutableHandle(objp),
-          MaybeRooted<PropertyResult, allowGC>::toMutableHandle(propp));
+      MOZ_ASSERT(!cx->isHelperThreadContext());
+      if constexpr (!allowGC) {
+        return false;
+      } else {
+        return LookupProperty(cx, proto, id, objp, propp);
+      }
     }
 
     current = &proto->template as<NativeObject>();
@@ -881,17 +829,27 @@ static MOZ_ALWAYS_INLINE bool LookupPropertyInline(
 
 inline bool ThrowIfNotConstructing(JSContext* cx, const CallArgs& args,
                                    const char* builtinName) {
-  if (args.isConstructing()) return true;
-  return JS_ReportErrorFlagsAndNumberASCII(cx, JSREPORT_ERROR, GetErrorMessage,
-                                           nullptr, JSMSG_BUILTIN_CTOR_NO_NEW,
-                                           builtinName);
+  if (args.isConstructing()) {
+    return true;
+  }
+  JS_ReportErrorNumberASCII(cx, GetErrorMessage, nullptr,
+                            JSMSG_BUILTIN_CTOR_NO_NEW, builtinName);
+  return false;
 }
 
 inline bool IsPackedArray(JSObject* obj) {
-  return obj->is<ArrayObject>() && !obj->hasLazyGroup() &&
-         !obj->group()->hasAllFlags(OBJECT_FLAG_NON_PACKED) &&
-         obj->as<ArrayObject>().getDenseInitializedLength() ==
-             obj->as<ArrayObject>().length();
+  if (!IsTypeInferenceEnabled()) {
+    return false;
+  }
+  if (!obj->is<ArrayObject>() || obj->hasLazyGroup()) {
+    return false;
+  }
+  AutoSweepObjectGroup sweep(obj->group());
+  if (obj->group()->hasAllFlags(sweep, OBJECT_FLAG_NON_PACKED)) {
+    return false;
+  }
+  return obj->as<ArrayObject>().getDenseInitializedLength() ==
+         obj->as<ArrayObject>().length();
 }
 
 }  // namespace js
